"""Compatibility module.

It can be necessary to load files generated by previous versions of cloudpickle
that rely on symbols being defined under the `cloudpickle.cloudpickle_fast`
namespace.

See: tests/test_backward_compat.py
"""
from . import cloudpickle


def __getattr__(name):
    return getattr(cloudpickle, name)
